@@ -1,8 +1,8 @@
 /* gdb_interface.c - core analysis suite
  *
  * Copyright (C) 1999, 2000, 2001, 2002 Mission Critical Linux, Inc.
- * Copyright (C) 2002, 2003, 2004, 2005, 2006, 2007 David Anderson
- * Copyright (C) 2002, 2003, 2004, 2005, 2006, 2007 Red Hat, Inc. All rights reserved.
+ * Copyright (C) 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010 David Anderson
+ * Copyright (C) 2002, 2003, 2004, 2005, 2006, 2007, 2008, 2009, 2010 Red Hat, Inc. All rights reserved.
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -786,6 +786,7 @@ int
 gdb_readmem_callback(ulong addr, void *buf, int len, int write)
 { 
 	char locbuf[SIZEOF_32BIT], *p1;
+	uint32_t *p2;
 	int memtype;
 
 	if (write)
@@ -830,8 +831,8 @@ gdb_readmem_callback(ulong addr, void *buf, int len, int write)
 		if (readmem(addr, memtype, locbuf, SIZEOF_32BIT,
                     "gdb_readmem_callback", RETURN_ON_ERROR)) {
 			*p1 = locbuf[0];
-			text_value_cache(addr, 
-				(uint32_t)*((uint32_t *)locbuf), 0);
+			p2 = (uint32_t *)locbuf;
+			text_value_cache(addr, *p2, 0);
 			return TRUE;
 		}
 		break;
