@@ -586,7 +586,7 @@ main_loop(void)
 
 	if (pc->flags & MINIMAL_MODE)
             error(NOTE, 
-		"minimal mode commands: log, dis, rd, sym, eval and exit\n\n");
+		"minimal mode commands: log, dis, rd, sym, eval, set and exit\n\n");
 
         pc->flags |= RUNTIME;
 
@@ -640,7 +640,7 @@ reattempt:
 #else
         			error(FATAL, XEN_HYPERVISOR_NOT_SUPPORTED);
 #endif
-			} else {
+			} else if (!(pc->flags & MINIMAL_MODE)) {
 				tt->refresh_task_table();
 				sort_context_array();
 			}
@@ -679,7 +679,7 @@ reattempt:
 	if (pc->flags & MINIMAL_MODE)
 		error(INFO, 
 		    "%s: command not available in minimal mode\n"
-		    "NOTE: minimal mode commands: log, dis, rd, sym, eval and exit\n",
+		    "NOTE: minimal mode commands: log, dis, rd, sym, eval, set and exit\n",
 			args[0]);
 	else
 		error(INFO, "command not found: %s\n", args[0]);
@@ -1346,6 +1346,8 @@ dump_program_context(void)
 		fprintf(fp, "%sIRQ_IN_USE", others ? "|" : "");
         if (pc->curcmd_flags & MODULE_TREE)
 		fprintf(fp, "%sMODULE_TREE", others ? "|" : "");
+        if (pc->curcmd_flags & IGNORE_ERRORS)
+		fprintf(fp, "%sIGNORE_ERRORS", others ? "|" : "");
 	fprintf(fp, ")\n");
 	fprintf(fp, "   curcmd_private: %llx\n", pc->curcmd_private); 
 	fprintf(fp, "       sigint_cnt: %d\n", pc->sigint_cnt);
